# Copyright 2021-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
set_global_routing_layer_adjustment * $::env(GRT_ADJUSTMENT)

set array [split $::env(GRT_LAYER_ADJUSTMENTS) " "]


set ::tech [[::ord::get_db] getTech]

set layer_names [list]
set layers [$::tech getLayers]
foreach layer $layers {
    if { [$layer getRoutingLevel] >= 1 } {
        lappend layer_names [$layer getName]
    }
}

set i 0
foreach adjustment $array {
    set layer_name [lindex $layer_names $i]
    set_global_routing_layer_adjustment $layer_name $adjustment
    incr i
}